import { defineComponent, PropType, h, ConcreteComponent, SetupContext, onMounted, reactive, watch, computed, mergeProps, toRaw, isRef, unref, Component, getCurrentInstance } from 'vue'
import { useRoute, RouteLocationNormalizedLoaded } from 'vue-router'
import { Meta, SchemaNodeList, SchemaNode, ComponentNodeList, NormalizedSchemaNode, NormalizedSchema } from '../schema'
import { PluginDriver, pluginDriver } from '../plugins'
import { useLoading, OpenLoading, CloseLoading } from './hooks/useLoading'
import useHook from './hooks/useHook'
import userParser from './hooks/userParser'
import { isFalse, mergeObjects, isString, normalizeSchema } from '../util'

export type RendererProps = NormalizedSchemaNode

export type ParserUtil = {
    openLoading: OpenLoading
    closeLoading: CloseLoading
    route: RouteLocationNormalizedLoaded
}

export const SchemaRenderer: Component<RendererProps> = defineComponent({
    name: 'Parser',
    props: {
        uniqueKey: {
            type: String,
            required: true,
        },
        type: {
            type: String,
            required: true,
        },
        label: {
            type: String,
        },
        meta: {
            type: Object as PropType<Meta>,
            default: () => ({})
        },
        items: {
            type: Array as PropType<SchemaNodeList>
        }
    },
    setup: (props, context) => {
        const route = useRoute()
        const {
            closeLoading,
            openLoading,
        } = useLoading()
        const parserUtil: ParserUtil = {
            closeLoading,
            openLoading,
            route,
        }

        const {component, props: parsedProps, events: componentEvents, isInit, children, methods} = userParser(props as NormalizedSchemaNode, context, parserUtil)
        const componentProps = computed( () => parsedProps && Object.entries(parsedProps).reduce((result, [key, value]) => Object.assign(result, {
            [key]: isRef(value) ? unref(value) : value
        }), {}))

        // 调用解析器
        const { callHook } = useHook(props as NormalizedSchemaNode)

        onMounted( () => {
            callHook('mounted')

            isFalse(isInit?.value) && openLoading()
        })

        watch( () => isInit?.value, (newV) => {
            newV && closeLoading()
        })

        // created
        callHook('created')

        // 将组件实例增加到ComponentManager里
        pluginDriver.addComponent(props.uniqueKey, getCurrentInstance()?.proxy)

        return {
            isInit,
            component,
            componentProps,
            componentEvents,
            children,
            closeLoading,
            openLoading,
            ...methods,
        }
    },
    render(){
        const {component, componentProps, componentEvents, isInit, children} = this

        const childrenComponents = children && (children as any).map( (item: SchemaNode | string) => {
            if(isString(item)) return item

            return h(SchemaRenderer as any, normalizeSchema(item))
        })


        if(isFalse(isInit)){
            return h('div')
        }

        return h((component as ConcreteComponent<RendererProps>), {...componentProps, ...componentEvents} as RendererProps, childrenComponents)
    }
})