// 解析器管理
import { SimpleManager } from "./common/simple";
import { BasePlugin, Util } from '../plugins'
import { PluginTypes } from '../constant'
import { RequestMeta } from '../actions/request'
import { Data, Meta, RestArguments } from '../schema'
import warn from "../util/warn";

export type ActionMeta = RequestMeta

export interface BaseAction {
    type: string
}

export type Action<T = Data> = T & BaseAction
export type ActionHandler = (action: Action, util?: Util, ...args: any[]) => any

export interface ActionPlugin extends BasePlugin {
    type: PluginTypes.Action,
    dependentMetas: string[],
    handler: ActionHandler
}

export class ActionManager extends SimpleManager{
    // 内部缓存对象
    protected list: ActionPlugin[]
    protected dependentMetas: Set<string>
    constructor(){
        super()
        this.name = 'Action'
        this.carReset = false
        this.list = []
        this.dependentMetas = new Set()
    }

    /**
     * 判断一个Meta是否为一个Action Meta
     *
     * @param   {Meta}     meta  待判定的Action Meta
     *
     * @return  {boolean}        结果
     */
    isActionMeta(meta: Meta): boolean{
        return !!(meta && Object.keys(meta).some( key => this.dependentMetas.has(key) ))
    }

    /**
     * 判断meta是否为某个特定类型的Action Meta
     *
     * @param   {string}  type  Action的类型
     * @param   {Meta}    meta  待判定的Action Meta
     *
     * @return  {boolean}       结果
     */
    isActionMetaByType(type: string, meta: Meta): boolean{
        const action = this.get(type) as ActionPlugin
        const {dependentMetas} = action || {}

        return !!(meta && dependentMetas && Object.keys(meta).some( key => dependentMetas.includes(key) ))
    }

    handleAction(meta: ActionMeta, util: Util, restArgs?: RestArguments): any{
        const action = this.getAction(meta)

        if(!action) return warn(`can not get acticon type of meta: `, meta)

        return this.callActionPlugin(action, {
            ...meta,
            type: action.name
        }, util, restArgs)
    }
    callActionPlugin(action: ActionPlugin, meta: Action<ActionMeta>, util: Util, restArgs: RestArguments = []){
        const { handler } = action

        return handler.call(null, meta, util, ...restArgs)
    }
    handleActionByType(type: string, meta: ActionMeta, util: Util, restArgs?: RestArguments){
        const action = this.get(type) as ActionPlugin

        return this.callActionPlugin(action, {
            ...meta,
            type
        }, util, restArgs)
    }
    getAction(meta: ActionMeta): ActionPlugin | undefined{
        const action = this.list.find( item => item.dependentMetas.some( key => key in meta ))

        return action
    }
    addDependentMetas(metaKeys: string[]){
        metaKeys.forEach( key => {
            this.dependentMetas.add(key)
        });
    }
    add(key: string, item: ActionPlugin): any{
        const isSuccess = super.add(key, item)

        if(isSuccess){
            this.list.push(item)
            this.addDependentMetas(item.dependentMetas)
        }
    }
}