import { Component, SetupContext, ComponentPublicInstance } from 'vue'
import { SchemaNode, ComponentNode, Meta, NormalizedSchemaNode, NormalizedSchema } from '../schema'
import BASE_PLUGINS from './base'
import { ParserManager, ParserPlugin } from '../manager/parser'
import { ActionManager, ActionPlugin, ActionMeta } from '../manager/action'
import { UIManager, UIConfig } from '../manager/ui'
import { Config, AppConfig } from '../engine'
import { ParserUtil } from '../components/renderer'
import { isUndefined } from '../util'
import { ParserTypes, PluginTypes, UITypes } from '../constant'
import { http, HttpRequestConfig } from '../util/http'
import warn from '../util/warn'
import { Plugin as PluginUI } from 'vue-seed-ui'
import { ComponentManager } from '../manager/component'

export interface BasePlugin {
    name: string,  // 插件名称
    type?: string, // 插件类型，默认为UI类型
}

// 解析组件辅助工具集
export interface PluginUtil {
    http: (options: HttpRequestConfig) => Promise<any>
    appConfig: AppConfig | null
    pluginDriver: PluginDriver
}

export type Util = PluginUtil & ParserUtil

export type ParseResult = ComponentNode | Promise<ComponentNode>

export type Plugin = ParserPlugin | ActionPlugin
type Plugins = Plugin[]
export type UserPlugins = Plugins | undefined

export class PluginDriver{
    options: Config | null
    plugins: Plugins
    parserManager: ParserManager
    actionManager: ActionManager
    uiManager: UIManager
    componentManager : ComponentManager
    // 辅助工具集合
    util: PluginUtil | null
    message?: PluginUI
    private static instance: PluginDriver

    constructor(){
        this.options = null
        this.util = null
        this.plugins = []

        // init managers
        this.uiManager = new UIManager()
        this.parserManager = new ParserManager()
        this.actionManager = new ActionManager()
        this.componentManager = new ComponentManager()

        this.init()
    }

    init(){
        this.distributePlugins(BASE_PLUGINS)
        this.initUtil()
    }

    initUtil(){
        this.util = {
            http: (options: HttpRequestConfig): Promise<any> => http(options).catch( err => this.message?.error(err.message || '服务错误') ),
            appConfig: null,
            pluginDriver: this
        }
    }

    setUtil(key: keyof PluginUtil, value: any){
        (this.util as Util)[key] = value
    }

    setOptions(options: Config){
        const {
            ui,
            plugins,
            appConfig
        } = options
        const {util} = this

        this.options = options

        plugins && this.distributePlugins(plugins)
        this.addUI(ui)
        util && (util.appConfig = appConfig)
    }

    // 根据插件类型，对插件进行分发
    distributePlugins(plugins: Plugins){
        const { plugins: installedPlugins } = this
        plugins.forEach( plugin => {
            const {type, name} = plugin

            if(installedPlugins.includes(plugin)) return warn(`${type} plugin ${name} has installed, please don't install it repeatedly`, plugin)

            installedPlugins.push(plugin)

            switch(type){
                case PluginTypes.Parser:
                    this.parserManager?.add(name, plugin)
                    break
                case PluginTypes.Action:
                    this.actionManager?.add(name, plugin as ActionPlugin)
                    break
                default:
                    warn(`unknown type of plugin ${name} with type ${type}`, plugin)
            }
        })
    }

    runHook(hookName: string, context: object, ...args: Array<keyof Plugin>){
        return this.plugins.reduce( (prevValue, plugin) => {
            const hook = plugin[hookName as keyof Plugin]

            if(!hook) return prevValue

            return typeof hook === 'function' ?  hook.apply(context, args) : hook
        }, undefined)
    }

    /**
     * 获取SchemaNode对应的解析器Parser
     *
     * @param   {SchemaNode}  options  [options description]
     *
     * @return  {[type]}               [return description]
     */
    getParser(options: NormalizedSchemaNode): ParserPlugin{
        const {type} = options as SchemaNode
        const parserManager = this.parserManager as ParserManager
        const parser = parserManager.get(type) || parserManager.get(ParserTypes.Common)

        return parser as ParserPlugin
    }

    getAppConfig(): AppConfig | null{
        const { options } = this

        if(options === null) return null

        const { appConfig } = options

        return appConfig
    }

    addComponent(key: string, componentInstance?: ComponentPublicInstance | null){
        componentInstance && this.componentManager.add(key, componentInstance)
    }

    getComponent(key: string){
        return this.componentManager.get(key)
    }

    parseSchema(options: NormalizedSchemaNode, context: SetupContext, util: ParserUtil): ComponentNode{
        const parserPlugin = this.getParser(options)
        const { parse, ui } = parserPlugin
        const pluginUtil: Util = Object.assign({}, this.util, util)
        const componentNode = parse.call(null, options, context, pluginUtil)
        let { component } = componentNode

        if(isUndefined(component)){
            componentNode.component = (typeof ui === 'string' ? this.getUI(ui) : ui) as Component
        }

        return componentNode
    }

    handleAction(meta: Meta, ...restArgs: any[]): any{
        const {actionManager, util} = this

        if(!actionManager || !actionManager.isActionMeta(meta)) return warn(`unknown action type with meta: `, meta)

        return actionManager.handleAction(meta as ActionMeta, util as Util, restArgs)
    }

    handleActionByType(type: string, meta: Meta): any{
        const {actionManager, util} = this

        if(!actionManager || !actionManager.isActionMetaByType(type, meta)) return warn(`unknown action type: ${type} with meta: `, meta)

        return actionManager.handleActionByType(type, meta as ActionMeta, util as Util)
    }

    getUI(type: string): Component{
        return this.uiManager?.get(type) as Component
    }

    addUI(ui: UIConfig){
        const {uiManager} = this

        uiManager?.addCaches(ui)

        // 设置完Message之后，要重新设置util的http
        if(UITypes.Message in ui){
            this.message = uiManager?.get(UITypes.Message) as PluginUI
        }
    }

    static getInstance(){
        return this.instance ? this.instance : (this.instance = new PluginDriver())
    }
}

export const pluginDriver = PluginDriver.getInstance()